from typing import Dict
from .api_base import ApiBase


class JournalEntries(ApiBase):
    """Class for Categories APIs."""

    GET_JOURNAL_ENTRIES = '/query?query=select * from JournalEntry STARTPOSITION {0} MAXRESULTS 1000'
    POST_JOURNAL_ENTRY = '/journalentry?minorversion=38'

    def get(self):
        """
        Get all JournalEntries
        :return: List of Dicts in JournalEntry Schema
        """
        return self._query_get_all('JournalEntry', JournalEntries.GET_JOURNAL_ENTRIES)

    def post(self, data: Dict):
        """
        Post JournalEntry (check, etc) to Quickbooks Online
        :param data: Dict in JournalEntry schema
        :return:
        """
        return self._post_request(data, JournalEntries.POST_JOURNAL_ENTRY)



from .api_base import ApiBase


class Accounts(ApiBase):
    """Class for Categories APIs."""

    GET_ACCOUNTS = '/query?query=select * from Account STARTPOSITION {0} MAXRESULTS 1000'

    def get(self):
        """Get a list of the existing Accounts in the Organization.

        Returns:
            List with dicts in Accounts schema.
        """
        return self._query_get_all('Account', Accounts.GET_ACCOUNTS)



from .qbosdk import QuickbooksOnlineSDK
from.exceptions import *

__all__ = [
    QuickbooksOnlineSDK,
    QuickbookOnlineSDKError,
    NotFoundClientError,
    UnauthorizedClientError,
    ExpiredTokenError,
    InvalidTokenError,
    NoPrivilegeError,
    WrongParamsError,
    NotFoundItemError,
    InternalServerError
]

name = "qbosdk"



from .api_base import ApiBase


class Preferences(ApiBase):
    """Class for Categories APIs."""

    GET_PREFERENCES = '/preferences'

    def get(self):
        """Get a list of the existing Preferences in the Organization.

        Returns:
            Dict in Preferences schema.
        """
        return self._get_request('Preferences', Preferences.GET_PREFERENCES)



from typing import Dict
from .api_base import ApiBase


class Purchases(ApiBase):
    """Class for Categories APIs."""

    GET_PURCHASES = '/query?query=select * from Purchase STARTPOSITION {0} MAXRESULTS 1000'
    POST_PURCHASE = '/purchase?minorversion=38'

    def get(self):
        """
        Get all Purchases
        :return: List of Dicts in Purchase Schema
        """
        return self._query_get_all('Purchase', Purchases.GET_PURCHASES)

    def post(self, data: Dict):
        """
        Post Purchase (check, etc) to Quickbooks Online
        :param data: Dict in Purchase schema
        :return:
        """
        return self._post_request(data, Purchases.POST_PURCHASE)



from .api_base import ApiBase


class Classes(ApiBase):
    """Class for Categories APIs."""

    GET_CLASSES = '/query?query=select * from Class STARTPOSITION {0} MAXRESULTS 1000'

    def get(self):
        """Get a list of the existing Classes in the Organization.

        Returns:
            List with dicts in Classes schema.
        """
        return self._query_get_all('Class', Classes.GET_CLASSES)



"""
Quickbooks online Python SDK
"""
import json
import base64
import requests
from future.moves.urllib.parse import urlencode

from .exceptions import *
from .apis import *


class QuickbooksOnlineSDK:
    """
    Quickbooks Online SDK
    """
    TOKEN_URL = 'https://oauth.platform.intuit.com/oauth2/v1/tokens/bearer'

    def __init__(self, client_id: str, client_secret: str,
                 refresh_token: str, realm_id: str, environment: str):
        """
        Initialize connection to Quickbooks Online
        :param client_id: Quickbooks online client_Id
        :param client_secret: Quickbooks online client_secret
        :param refresh_token: Quickbooks online refresh_token
        :param realm_id: Quickbooks onliine realm / company id
        :param environment: production or sandbox
        """
        # Initializing variables
        self.__client_id = client_id
        self.__client_secret = client_secret
        self.refresh_token = refresh_token

        if environment.lower() == 'production':
            self.__base_url = 'https://quickbooks.api.intuit.com/v3/company/{0}'.format(realm_id)
            self.web_app_url = 'https://app.qbo.intuit.com'
        elif environment.lower() == 'sandbox':
            self.__base_url = 'https://sandbox-quickbooks.api.intuit.com/v3/company/{0}'.format(realm_id)
            self.web_app_url = 'https://app.sandbox.qbo.intuit.com'
        else:
            raise ValueError('environment can only be prodcution / sandbox')

        self.__access_token = None

        self.accounts = Accounts()
        self.departments = Departments()
        self.classes = Classes()
        self.employees = Employees()
        self.preferences = Preferences()
        self.exchangeRates = ExchangeRates()
        self.purchases = Purchases()
        self.journal_entries = JournalEntries()
        self.attachments = Attachments()

        self.update_server_url()
        self.update_access_token()

    def update_server_url(self):
        """
        Update the server url in all API objects.
        """
        base_url = self.__base_url

        self.accounts.set_server_url(base_url)
        self.departments.set_server_url(base_url)
        self.classes.set_server_url(base_url)
        self.employees.set_server_url(base_url)
        self.preferences.set_server_url(base_url)
        self.exchangeRates.set_server_url(base_url)
        self.purchases.set_server_url(base_url)
        self.journal_entries.set_server_url(base_url)
        self.attachments.set_server_url(base_url)

    def update_access_token(self):
        """
        Update the access token and change it in all API objects.
        """
        self.__get_access_token()
        access_token = self.__access_token

        self.accounts.change_access_token(access_token)
        self.departments.change_access_token(access_token)
        self.classes.change_access_token(access_token)
        self.employees.change_access_token(access_token)
        self.preferences.change_access_token(access_token)
        self.exchangeRates.change_access_token(access_token)
        self.purchases.change_access_token(access_token)
        self.journal_entries.change_access_token(access_token)
        self.attachments.change_access_token(access_token)

    def __get_access_token(self):
        """Get the access token using a HTTP post.

        Returns:
            A new access token.
        """

        api_data = {
            'grant_type': 'refresh_token',
            'refresh_token': self.refresh_token
        }

        auth = '{0}:{1}'.format(self.__client_id, self.__client_secret)
        auth = base64.b64encode(auth.encode('utf-8'))

        request_header = {
            'Accept': 'application/json',
            'Content-type': 'application/x-www-form-urlencoded',
            'Authorization': 'Basic {0}'.format(
                str(auth.decode())
            )
        }

        token_url = QuickbooksOnlineSDK.TOKEN_URL.format(self.__base_url)
        response = requests.post(url=token_url, data=urlencode(api_data), headers=request_header)

        if response.status_code == 200:
            auth = json.loads(response.text)
            self.__access_token = auth['access_token']
            self.refresh_token = auth['refresh_token']

        elif response.status_code == 400:
            raise WrongParamsError('Somthing wrong with the request body', response.text)

        elif response.status_code == 401:
            raise UnauthorizedClientError('Wrong client secret or/and refresh token', response.text)

        elif response.status_code == 404:
            raise NotFoundClientError('Client ID doesn\'t exist', response.text)

        elif response.status_code == 500:
            raise InternalServerError('Internal server error', response.text)

        else:
            raise QuickbooksOnlineSDK('Error: {0}'.format(response.status_code), response.text)



class QuickbookOnlineSDKError(Exception):
    """The base exception class for QBOSDK.

    Parameters:
        msg (str): Short description of the error.
        response: Error response from the API call.
    """

    def __init__(self, msg, response=None):
        super(QuickbookOnlineSDKError, self).__init__(msg)
        self.message = msg
        self.response = response

    def __str__(self):
        return repr(self.message)


class NotFoundClientError(QuickbookOnlineSDKError):
    """Client not found OAuth2 authorization, 404 error."""
    pass


class UnauthorizedClientError(QuickbookOnlineSDKError):
    """Wrong client secret and/or refresh token, 401 error."""
    pass


class ExpiredTokenError(QuickbookOnlineSDKError):
    """Expired (old) access token, 498 error."""
    pass


class InvalidTokenError(QuickbookOnlineSDKError):
    """Wrong/non-existing access token, 401 error."""
    pass


class NoPrivilegeError(QuickbookOnlineSDKError):
    """The user has insufficient privilege, 403 error."""
    pass


class WrongParamsError(QuickbookOnlineSDKError):
    """Some of the parameters (HTTP params or request body) are wrong, 400 error."""
    pass


class NotFoundItemError(QuickbookOnlineSDKError):
    """Not found the item from URL, 404 error."""
    pass


class InternalServerError(QuickbookOnlineSDKError):
    """The rest QBOSDK errors, 500 error."""
    pass



import json
import textwrap
from typing import Dict

from .api_base import ApiBase


class Attachments(ApiBase):
    """Class for Categories APIs."""

    GET_ATTACHABLES = '/query?query=select * from Attachable STARTPOSITION {0} MAXRESULTS 1000'
    POST_ATTACHABLE = '/attachable'
    POST_FILE = '/upload'

    def get(self):
        """
        Get all Attachables
        :return: List of Dicts in Attachable Schema
        """
        return self._query_get_all('Attachable', Attachments.GET_ATTACHABLES)

    def post(self, ref_id: str, ref_type: str, content, file_name):
        """
        Post Attachable (check, etc) to Quickbooks Online
        :param file_name: name of the file
        :param content: attachment content
        :param ref_type: type of object
        :param ref_id: object id
        :return: Dict in attachable schema
        """
        file = self.__upload_file(content, file_name)
        if file:
            attachable_ref = [
                {
                    'EntityRef': {
                        'type': ref_type,
                        'value': ref_id
                    }
                }
            ]
            file['AttachableRef'] = attachable_ref
        return self._post_request(file, Attachments.POST_ATTACHABLE)

    def __upload_file(self, content, file_name: str) -> Dict:
        """
        Upload file to Quickbooks
        :param content: Content in base64
        :param file_name: name of the file
        :return: Returns file object as Dict
        """
        binary_data = content

        boundary = 'YOjcLaTlykb6OxfYJx4O07j1MweeMFem'

        content_type = self.__get_content_type(file_name)

        body = {
            'Content-Type': content_type
        }

        if content_type:
            request_body = textwrap.dedent(
                """
                --%s
                Content-Disposition: form-data; name="file_metadata_01";
                Content-Type: application/json
                %s
                --%s
                Content-Disposition: form-data; name="file_content_01"; filename="%s"
                Content-Type: %s
                Content-Transfer-Encoding: base64

                %s

                --%s--
                """
            ) % (boundary, json.dumps(body), boundary, file_name, content_type, binary_data, boundary)

            return self._post_file(str(request_body), Attachments.POST_FILE)

        return {}

    @staticmethod
    def __get_content_type(file_name: str) -> str or None:
        """
        Gets content type of supported file types
        :param file_name: name of the file
        :return: content-type or None
        """
        extension = file_name.split('.')[-1]

        content_types: Dict = {
            'ai': 'application/postscript',
            'csv': 'text/csv',
            'doc': 'application/msword',
            'docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
            'eps': 'application/postscript',
            'gif': 'image/gif',
            'jpeg': 'image/jpeg',
            'jpg': 'image/jpg',
            'png': 'image/png',
            'ods': 'application/vnd.oasis.opendocument.spreadsheet',
            'pdf': 'application/pdf',
            'rtf': 'text/rtf',
            'tif': 'image/tiff',
            'txt': 'text/plain',
            'xls': 'application/vnd.ms-excel',
            'xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
            'xml': 'text/xml'
        }

        return content_types.get(extension.lower())



from datetime import datetime, timedelta
from .api_base import ApiBase


class ExchangeRates(ApiBase):
    """Class for Categories APIs."""

    GET_EXCHANGE_RATES = "/query?query=select * from ExchangeRate where AsOfDate = '{0}' STARTPOSITION " \
                         "{1} MAXRESULTS 1000"

    def get(self, as_of_date: str = None):
        """
        Get all the exchange rates
        :param as_of_date: date to get rates for (1 day prior if left empty)
        :return: List of Dicts for exchange rates
        """
        if not as_of_date:
            as_of_date = (datetime.now() - timedelta(days=1)).strftime("%Y-%m-%d")

        ExchangeRates.GET_EXCHANGE_RATES = ExchangeRates.GET_EXCHANGE_RATES.format(as_of_date, '{0}')
        return self._query_get_all('ExchangeRate', ExchangeRates.GET_EXCHANGE_RATES)



from .api_base import ApiBase


class Departments(ApiBase):
    """Class for Categories APIs."""

    GET_DEPARTMENTS = '/query?query=select * from Department STARTPOSITION {0} MAXRESULTS 1000'

    def get(self):
        """Get a list of the existing Departments in the Organization.

        Returns:
            List with dicts in Departments schema.
        """
        return self._query_get_all('Department', Departments.GET_DEPARTMENTS)



from .accounts import Accounts
from .departments import Departments
from .classes import Classes
from .employees import Employees
from .preferences import Preferences
from .exchange_rates import ExchangeRates
from .purchases import Purchases
from .journal_entries import JournalEntries
from .attachments import Attachments


__all_ = [
    'Accounts',
    'Departments',
    'Classes',
    'Employees',
    'Preferences',
    'ExchangeRates',
    'Purchases',
    'JournalEntries',
    'Attachments'
]



from .api_base import ApiBase


class Employees(ApiBase):
    """Class for Categories APIs."""

    GET_EMPLOYEES = '/query?query=select * from Employee STARTPOSITION {0} MAXRESULTS 1000'

    def get(self):
        """Get a list of the existing Employees in the Organization.

        Returns:
            List with dicts in Employees schema.
        """
        return self._query_get_all('Employee', Employees.GET_EMPLOYEES)
